/* 
**	GKDeReg - Custom uninstaller for the Freeware GK installation program
**			This program removes registry entries (The piece GK doesn't 
**			have in the free program.) It also closes any running windows
**			that belong to a class that is the same as the Application,
**			or have the same name as the Application.
************************************************************************************
**  To use this program, your application must conform to these standard Registry
**  conventions:
**
**  AutoStart Value has the same name as the Application in:
**  HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\Run
**  or
**  HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Run
**
**  All application data is in the following Registry hive:
**  HKEY_CURRENT_USER\\Software\\CompanyName\\AppName
**
**  Command line might look like this:
**  GKDeReg CompanyABC Application123 Y Y Y
**
**  The first "Y" indicates that the Company key is to be removed. ("N" means leave it)
**  The second "Y" indicates that the Application AutoStart key is to be removed. ("N" means leave it)
**  The third "Y" indicates that the Application should be stopped now. ("N" means leave it)
**
**  This command line would remove the following keys:
**  HKEY_CURRENT_USER\\Software\\CompanyABC\\Application123
**  HKEY_CURRENT_USER\\Software\\CompanyABC
**
**  And any values named "Application123" will be removed from the following keys:
**  HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\Run
**  HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Run
**
**  Also, any windows named "Application123", (or belonging to a window 
**  class called "Application123"), are closed.
**
************************************************************************************
**	File:    GKDeReg.c
**	Author:  John Fitzgibbon (fitz@jfitz.com http://www.jfitz.com)
**	Version: 1.0
**	Date:    11/27/1999
**
**	This program and code is FREEWARE, but please give credit where credit is due.
**
**	Compiled with Visual C++ 4.0. 
**
**  Revision History
**  11/27/1999 1.0 Original Version 
**
*/

#include <windows.h>  // main Windows header
#include <string.h>

char CompanyName[_MAX_PATH];
char AppName[_MAX_PATH];
char KeyPath[_MAX_PATH];
char msg[_MAX_PATH+256];

const char RunSubKeyStr[] = "Software\\Microsoft\\Windows\\CurrentVersion\\Run";

//
// This function uses strtok to return parameters separated by spaces, 
// and/or enclosed in double quotes.
//
char * cmdline_strtok(char * str)
{
	char * param;

	param = strtok(str," ");
	if (param && (param[0] == '"'))
		param = strtok(NULL,"\"");

	return param;
}

//
// This function deletes the specified key.
//
// All subkeys and values are also deleted. This function does what RegDeleteKey
// does for Windows 95/98 and what it should do, (but doesn't), for NT.
//
LONG RegReallyDeleteKey(
    HKEY  hKey,	// handle of open key 
    LPCTSTR  lpSubKey 	// address of name of subkey to delete 
   )
{
	//
	// Make most of the local variables static to conserve stack space.
	// This function uses recursion which could be quite stack-intensive
	// if the sub keys are nested very deeply.
	//
	HKEY hSubKey;			// needs to be preserved after recursive call - not static
	char Name[_MAX_PATH];	// needs to be preserved after recursive call - not static
	LONG RetVal;			// needs to be preserved after recursive call - not static
	static DWORD NameSize;
	static char Class[_MAX_PATH];
	static DWORD ClassSize;
	static FILETIME LastWriteTime;
    static DWORD Type;

	// MessageBox(NULL,lpSubKey,"RegDeleteKey", MB_OK);
	if (RegOpenKeyEx(hKey, lpSubKey, 0, KEY_ALL_ACCESS, &hSubKey) == ERROR_SUCCESS)
	{
		//
		// Enumerate the sub keys and blow them away with recursive calls to this function.
		//
		RetVal = ERROR_SUCCESS;
		while (RetVal == ERROR_SUCCESS)
		{
			NameSize = ClassSize = _MAX_PATH;
			if ((RetVal = RegEnumKeyEx(
					    hSubKey,	// handle of key to enumerate 
					    0,	// index of subkey to enumerate 
					    Name,	// address of buffer for subkey name 
					    &NameSize,	// address for size of subkey buffer 
					    NULL,		// reserved 
					    Class,	// address of buffer for class string 
					    &ClassSize,	// address for size of class buffer 
						&LastWriteTime 	// address for time key last written to 
					   )) == ERROR_SUCCESS)
			{
				RegReallyDeleteKey(hSubKey, Name);
				// if (RegReallyDeleteKey(hSubKey, Name) != ERROR_SUCCESS)
				//	MessageBox(Name,"RegDeleteKey Failed",MB_ICONEXCLAMATION|MB_OK);
				// Note: remember all your static variables are potentially 
				//       changed by the time you get here
			}
			// else if (RetVal != ERROR_NO_MORE_ITEMS)
			//	MessageBox(Name,"RegEnumKeyEx Failed",MB_OK);
		}
		//
		// Enumerate the values and blow them away.
		//
		RetVal = ERROR_SUCCESS;
		while (RetVal == ERROR_SUCCESS)
		{
			NameSize = _MAX_PATH;
			if ((RetVal = RegEnumValue(
					    hSubKey,	// handle of key to enumerate 
					    0,	// index of subkey to enumerate 
					    Name,	// address of buffer for subkey name 
					    &NameSize,	// address for size of subkey buffer 
					    NULL,		// reserved
						&Type,	// address of buffer for type code 
						NULL,	// address of buffer for value data 
						NULL	// address for size of data buffer 
					   )) == ERROR_SUCCESS)
			{
				RegDeleteValue(hSubKey, Name);
				// if (RegDeleteValue(hSubKey, Name) != ERROR_SUCCESS)
				//	MessageBox(Name,"RegDeleteValue Failed",MB_ICONEXCLAMATION|MB_OK);
			}
			// else if (RetVal != ERROR_NO_MORE_ITEMS)
			//	MessageBox(Name,"RegEnumValue Failed",MB_OK);
		}
		RegCloseKey (hSubKey);

	}
	// else
	//	MessageBox(lpSubKey,"RegOpenKeyEx Failed",MB_OK|MB_ICONEXCLAMATION);

	//
	// Note: we've ignored error checking throughout.
	// If anything failed to delete, then the following RegDeleteKey will fail on NT.
	// On Windows 95/98 it may/may not fail. (It probably will fail, because there 
	// shouldn't have been any errors with the prior deletes.)
	//
	return RegDeleteKey(hKey, lpSubKey);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR lpszCmdLine, int nCmdShow)
{
	HWND hWnd;			// Handle of any running RAMpage Window
	LRESULT msgr;		// Message Result
	HKEY hKey;			// Handle of Registry Key
	int attempts = 0;
	BOOL RemoveCompany = FALSE;
	BOOL RemoveAutoStart = FALSE;
	BOOL StopApp = FALSE;
	char * param;

	//
	// Read the Company Name
	//
	param = cmdline_strtok(lpszCmdLine);
	if (!param || !strlen(param))
		return 0;
	strcpy(CompanyName,param);
	// MessageBox(NULL, CompanyName,"DEBUG",MB_ICONINFORMATION|MB_OK);

	//
	// Read the Application Name
	//
	param = cmdline_strtok(NULL);
	if (!param || !strlen(param))
		return 0;
	strcpy(AppName,param);
	// MessageBox(NULL, AppName,"DEBUG",MB_ICONINFORMATION|MB_OK);

	//
	// Read the "Remove Company Name" flag
	//
	param = cmdline_strtok(NULL);
	if (!param || !strlen(param) || ((param[0] != 'Y') && (param[0] != 'N')))
		return 0;
	if (param[0] == 'Y')
		RemoveCompany = TRUE;

	//
	// Read the "Remove Application AutoStart" flag
	//
	param = cmdline_strtok(NULL);
	if (!param || !strlen(param) || ((param[0] != 'Y') && (param[0] != 'N')))
		return 0;
	if (param[0] == 'Y')
		RemoveAutoStart = TRUE;

	//
	// Read the "Stop Application" flag
	//
	param = cmdline_strtok(NULL);
	if (!param || !strlen(param) || ((param[0] != 'Y') && (param[0] != 'N')))
		return 0;
	if (param[0] == 'Y')
		StopApp = TRUE;

	//
	// Build the appropriate key path for the Company or Application then blow it away
	//
	strcpy(KeyPath,"Software\\");
	strcat(KeyPath,CompanyName);

	if (!RemoveCompany)		// Leave the company and just blow away the app.
	{
		strcat(KeyPath,"\\");
		strcat(KeyPath,AppName);
	}

	//
	// Delete the key
	//
	if (RegReallyDeleteKey(HKEY_CURRENT_USER, KeyPath) != ERROR_SUCCESS)
	{
		strcpy(msg,"WARNING: Registry entries could not be removed.\n\n");
		strcat(msg,"The uninstall process will continue anyway.\n\n");
		strcat(msg,"The failed Registry key is:\n\nHKEY_CURRENT_USER\\");
		strcat(msg,KeyPath);
		MessageBox(NULL,msg,"Unable to remove Registry Key",MB_OK);
	}

	//
	// Get rid of the Auto Start registry entry
	//
	if (RemoveAutoStart)
	{
		//
		// Note: Try to delete the value from both HKEY_LOCAL_MACHINE and HKEY_CURRENT_USER
		// No error message is given if this fails, as there may be no entry to begin with.
		//
		if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, RunSubKeyStr, 0, KEY_SET_VALUE, &hKey) == ERROR_SUCCESS)
		{
			RegDeleteValue(hKey, AppName);
			RegCloseKey(hKey);
		}
		if (RegOpenKeyEx(HKEY_CURRENT_USER, RunSubKeyStr, 0, KEY_SET_VALUE, &hKey) == ERROR_SUCCESS)
		{
			RegDeleteValue(hKey, AppName);
			RegCloseKey(hKey);
		}
	}

	if (StopApp)
	{
		//
		// Close windows with class or window names matching the application name
		// - This is a little dangerous! Make sure the app name is unique if
		//   you use this option!
		//
		hWnd = FindWindow (AppName,NULL);
		while ((hWnd != NULL) && (attempts < 10))
		{
			msgr = SendMessage(
				hWnd,	// handle of destination window
				WM_CLOSE,	// message to send
				(WPARAM) 0,	// first message parameter
				(LPARAM) 0	// second message parameter
				);
			Sleep(500); // Give the app time to stop before continuing
			attempts++;
			hWnd = FindWindow (AppName,NULL);
		}
		//
		// Close windows with window names matching the application name
		//
		hWnd = FindWindow (NULL,AppName);
		while ((hWnd != NULL) && (attempts < 10))
		{
			msgr = SendMessage(
				hWnd,	// handle of destination window
				WM_CLOSE,	// message to send
				(WPARAM) 0,	// first message parameter
				(LPARAM) 0	// second message parameter
				);
			Sleep(500); // Give the app time to stop before continuing
			attempts++;
			hWnd = FindWindow (NULL,AppName);
		}
	}

	return 0;
}

