/* 
**	RAMpage - System Tray Free Memory display
**
**	File:    RAMPage.c
**	Author:  John Fitzgibbon (fitz@jfitz.com http://www.jfitz.com)
**	Version: 1.5
**	Date:    06/05/2000
**	Credits: Based on system tray example Tray42 by Michael T. Smith / R.A.M. Technology.
**	         (aa529@chebucto.ns.ca>
**
**	This program and code is FREEWARE, but please give credit where credit is due.
**
**	Compiled with Visual C++ 4.0. 
**  To optimize the executable size, choose "ignore default libraries" in the linker
**  settings and explicitly specify the following libraries:
**  kernel32.lib user32.lib libc.lib shell32.lib gdi32.lib
**
**  Revision History
**  08/30/1999 1.0 - Original Version 
**  11/02/1999 1.1 - Fixed NT display, (converts DIB to DDB). 
**                 - Added Luminous numbers.
**  11/12/1999 1.2 - Optimized the executable size by explicitly specifying libraries.
**                   (previous versions used all bloated debug libraries)
**                   Executable was 101K, now 60K. Memory footprint was 190K now about 180K
**                 - Also made one or two minor optimizations to the code.
**                 - Now rounds up/down to nearest MByte. (was truncating).
**                 - Added "B" option to display "Details" box in Bytes (default is MBytes).
**  11/27/1999 1.3 - Added "T" option to allow specification of Minimum Memory Threshold in MBytes.
**                   (Memory is freed automatically when free RAM falls below this value.)
**                 - Added code to limit how often automatic freeing is performed,
**                   (uses a "doubling" algorithm, with a 10 minute upper limit, to prevent 
**                    automatic freeing from running over and over when no memory can be freed).
**                 - Various optimizations: packed color palette and digit bitmaps, 
**                   "const" identifiers, removed floating point ops, minimized library calls,
**                   removed unnecessary code, data, etc, etc.
**                 - Fixed memory leak with "Change Display" popup menu option
**                   (option is no longer available, other options checked for leaks)
**  02/19/2000 1.4 - Added "Hidden" compilation option, (allows program to run without tray icon).
**                   Running hidden means automated free only, but memory footprint is smaller.
**                 - Added "O" option to "run Once". This does a free, then exits immediately.
**                   (Threshold, display and update intervals have no effect in this mode.)
**                   (Makes most sense to use this option with the "hidden" version.)
**                 - Added "S", (Sound) option. (Beeps when freeing memory automatically.)
**                 - Added memory defragmentation option. "D=Y" means always defragment memory.
**                   "D=N" means never defrag. Default is to defrag when freeing < 1/3 total mem.
**                   (Default should mean faster frees when silly users try to free too much mem.)
**                 - Added "P", (Path) option. If included the "Help/Config" menu item runs 
**                   RAMpageConfig.exe using the specified Path. (Otherwise, the existing 
**                   Help/About box is displayed.) Note: Path MUST be the last parameter.
**                 - Changed color definition for "white" from RGB=(0xFF,0xFF,0xFF) to 
**                   (0xF0,0xF0,0xF0) to fix NT display of white as dark gray. Note: 256 color
**                   display on NT still shows up as dark gray. Workaround is to use more colors
**                   or use the luminous green digit display type.
**                 - Changed from "top down" to "bottom up" bitmap specification - some users
**                   report upside down tray icons with "top down" bitmaps.
**                 - Automatic freeing is never called if Threshold=0. (Prior version would do
**                   automatic freeing if free memory was actually = 0 bytes).
**  06/05/2000 1.5 - Added graphic display type G=N
**                   N=1 simple guage. height of bar rises and falls with memory
**                   N=2 bar graph. bar moving from right to left tracks memory
**                   N=3 line graph. line moving from right to left tracks memory.
**                   If "N" is not 1, 2 or 3, then the simple guage is used.
**                 - Added "Always", ("A"), option. Frees memory even if last attempt failed.
**                 - Moved all language specific code into the header and resource file.
**                 - Now uses optimized malloc + memset. Reduces executable size by about 18K
**                   and memory footprint by about 10K. Thanks to Vladimir Sorokosh.
**                 - Removed the use of a single bitmap for foreground/background bmaps to fix
**                   display problems on some configurations. Thanks again to Vladimir Sorokosh.
**  05/01/2001 1.6 - Added "reach target" option R=X (Attempts to reach a target of X MB free)
**                   The "R" option is added to support the "rr" utility, (run and release).
**                   This utility can execute a program then trigger RAMpage by sending it a
**                   double click message.
**                 - Switched to lcc (freeware) compiler. (This reduces the code size.)
*/

#include <windows.h>  // main Windows header
#include <windowsx.h> // message cracker macros, etc.
#include <shellapi.h> // Shell32 api
#include <stdio.h>
#include <malloc.h>
#include "RAMpage.h"

#ifdef _RAMPAGE_RESMON
#include "k32exp.c"

// From 16-bit windows.h
#define GFSR_SYSTEMRESOURCES 0
#define GFSR_GDIRESOURCES 1
#define GFSR_USERRESOURCES 2

WORD GetFreeSystemResources(WORD type);
void GetResourceInfo();

WORD lSysRes;
WORD lGdiRes;
WORD lUsrRes;

void GetResourceInfo()
{
	lSysRes  = GetFreeSystemResources(GFSR_SYSTEMRESOURCES);
	lGdiRes  = GetFreeSystemResources(GFSR_GDIRESOURCES);
	lUsrRes = GetFreeSystemResources(GFSR_USERRESOURCES);
}
																	   
//We need this pragma to make a compiler even don't think about to modify a code in release mode
#pragma optimize ("", off) 
WORD GetFreeSystemResources(WORD type)
{
//SAFE macro just returns 99 if something goes wrong (say under NT)
#define DUMMY_RES 99
#define	SAFE(a) if(!(a)) return DUMMY_RES
	
	//Decarations  "c" style
	typedef HINSTANCE (WINAPI* LOADLIB16)(char*);
	typedef LONG (WINAPI* GETPROC16)(HINSTANCE, LPCSTR);
	typedef BOOL (WINAPI* FREELIB16)(HINSTANCE);
	FARPROC QT_Thunk;

	LOADLIB16 pLoadLib16;
	GETPROC16 pGetProc16;
	FREELIB16 pFreeLib16;
	HINSTANCE hInst;
	DWORD pGetRes;
	WORD user_fsr;
	WORD _type;
	HMODULE hKernel;

	//Usuall Windows API calls
	SAFE( hKernel = GetModuleHandle("KERNEL32"));
	SAFE( QT_Thunk = GetProcAddress(hKernel, "QT_Thunk"));

	//Andrew Schulman's 
	SAFE( pLoadLib16 = (LOADLIB16) GetK32ProcAddress(LOADLIBRARY16_ORD));
	SAFE( pGetProc16 = (GETPROC16) GetK32ProcAddress(GETPROCADDRESS16_ORD));
	SAFE( pFreeLib16 = (FREELIB16) GetK32ProcAddress(FREELIBRARY16_ORD));

	//Usuall for "c"-style calls - by pointer to function 
	SAFE( hInst = (*pLoadLib16)("user"));
		(*pFreeLib16)(hInst);
	SAFE( pGetRes = (*pGetProc16)(hInst,"GetFreeSystemResources"));

	//How to call 16-bit dll API from 32-bit .exe using QT_Thunk function
	/*Actually it is  

			__asm mov     edx, *pGetRes
			return (*QT_Thunk)(type);

			but compiler may produce a differnt code.
			To make QT_Thunk work we need exact code bellow
	*/

	_type = type;//for inline asm only
	if (pGetRes)
		_asm
		{  	
			push    _type
			mov     edx, [pGetRes] 
			call    QT_Thunk 
			mov     [user_fsr], ax
		}
	return user_fsr;


	/*There is another approach in MSVC documentation - "Thunk Compiler" chapter.
	I tried to learn about "Thunk Compiler" several times  and finally understood nothing.*/

#undef DUMMY_RES
#undef SAFE
}
#pragma optimize ("", on)

#endif

#define MBYTE ((DWORD)((DWORD)1024 * (DWORD)1024))
#define HALF_MBYTE ((DWORD)((DWORD)512 * (DWORD)1024))

// #define _RAMPAGE_HIDDEN		// This is now done in the project configuration

#ifndef _RAMPAGE_HIDDEN

#define UWM_SYSTRAY (WM_USER + 1) // Sent to us by the systray

#define BMP_RED (BYTE)0x66
#define BMP_GREEN (BYTE)0x44
#define BMP_DARK_GREEN (BYTE)0x54
#define BMP_YELLOW (BYTE)0x77
#define BMP_WHITE (BYTE)0x22
#define BMP_BLACK (BYTE)0x00

#endif

// #define _RAMPAGE_DEBUG
// #define _RAMPAGE_FREE_DEBUG
// #define _RAMPAGE_COLOR_DEBUG

#ifdef _RAMPAGE_FREE_DEBUG
char DebugStr[256];
#endif

#ifdef _RAMPAGE_DEBUG
// in debug mode, set bTesting to TRUE for an icon test display
BOOL bTesting = FALSE;

// test_num cycles from 0 to 10 each time the update timer fires.
// On cycle 0 a "test color card" is displayed.
// The following cycles display the numbers 000, 111, 222 ... 999
int test_num = 0;

BOOL bDetailsInBytes = TRUE;
#else

#ifndef _RAMPAGE_HIDDEN

BOOL bDetailsInBytes = FALSE;

#endif

#endif

const char classname[] = "RAMpage"; // Window class name

#ifndef _RAMPAGE_HIDDEN

#ifdef _RAMPAGE_GERMAN
#include "RAMpage_German.h"
#else
#ifdef _RAMPAGE_SPANISH
#include "RAMpage_Spanish.h"
#else
#ifdef _RAMPAGE_FRENCH
#include "RAMpage_French.h"
#else
#ifdef _RAMPAGE_DUTCH
#include "RAMpage_Dutch.h"
#else
#ifdef _RAMPAGE_TURKISH
#include "RAMpage_Turkish.h"
#else
#include "RAMpage_English.h"
#endif
#endif
#endif
#endif
#endif

#ifdef _RAMPAGE_RESMON
const char DetailsMessage[] = 
//					"%s\n\n"
					"%s%s%6d%c"
					"\n%s%s%6d%c"
					"\n\n%s%s%6d%c "
					"\n%s%s%6d%c"
					"\n\n%s%s%6d%c"
					"\n%s%s%6d%c"
//					"\n\n%s"
					"\n\n%s%s%6d%%"
					"\n%s%s%6d%%"
					"\n%s%s%6d%%";
#else
const char DetailsMessage[] = 
					"%s%s%6d%c"
					"\n%s%s%6d%c"
					"\n\n%s%s%6d%c "
					"\n%s%s%6d%c"
					"\n\n%s%s%6d%c"
					"\n%s%s%6d%c";
#endif

//
// We need a local type to store our eight color palette, which we cast to
// a BITMAPINFO when required. (BITMAPINFO has only one element in its RGB array.)
//
typedef struct tagLOCALBITMAPINFO { // bmi 
   BITMAPINFOHEADER bmiHeader; 
   RGBQUAD          bmiColors[8]; 
} LOCALBITMAPINFO; 

//
// localbmi stores the device independant bitmap information that is used to
// create the device dependant icon.
//
const LOCALBITMAPINFO localbmi = {
	{
		sizeof(BITMAPINFOHEADER),	// biSize
		16,							// biWidth 
		16,							// biHeight (negative means "top down" bitmap - doesn't work for some users)
		1,							// biPlanes
		4,							// biBitCount
		BI_RGB,						// biCompression
		0,							// biSizeImage
		0,							// biXPelsPerMeter
		0,							// biYPelsPerMeter
		8,							// biClrUsed
		8							// biClrImportant
	},
	{	//
		// What follows is RGB data for an 8 color subset of a standard 16 color palette.
		//
		// Note: The color positioning is significant.
		// Colors 0 - 3, (black, light gray, white and dark gray) are used
		// for the "normal" display. Keeping them as the low order colors
		// means the normal digit bitmaps can be packed 2 bits per pixel.
		// "Luminous" digits are essentially monochrome, and are packed
		// 1 bit per pixel in the digit bitmaps. Positioning Green at 4
		// means the bitmap can be unpacked with a simple shift (0001 -> 0100)
		// Positioning Green and White as colors 4 and 2 allows us to change 
		// from green digits to white digits with a simple right shift 
		// (i.e. a divide by 2)
		//
		{0x00,0x00,0x00,0x00}, // black			0->0
		{0xC0,0xC0,0xC0,0x00}, // light gray	7->1
		{0xF0,0xF0,0xF0,0x00}, // white			F->2  // True white, (0xFF), was displaying dark gray on NT
		{0x80,0x80,0x80,0x00}, // dark gray		8->3
		{0x00,0xFF,0x00,0x00}, // green			A->4
		{0x00,0xFF,0x00,0x00}, // dark green	2->5
		{0x00,0x00,0xFF,0x00}, // red			9->6
		{0x00,0xFF,0xFF,0x00}  // yellow		B->7
	}
};

//
// Bitmaps for digits displayed in the systray icon.
//

// Note: The comments are just a guide to help me map
// from an old bitmapping scheme to the new, more densely packed, format
//
// 0 = 0000, 1 = 0005, 2 = 0050, 3 = 0055
// 4 = 0500, 5 = 0505, 6 = 0550, 7 = 0555
// 8 = 5000, 9 = 5005, A = 5050, B = 5055
// C = 5500, D = 5505, E = 5550, F = 5555
//

const BYTE Numbers_Luminous[40] = {
	// "0" Bitmap
	0xF9, // 0x55, 0x55,	0x50, 0x05,
	0x99, // 0x50, 0x05,	0x50, 0x05,
	0x99, // 0x50, 0x05,	0x50, 0x05,
	0x9F, // 0x50, 0x05,	0x55, 0x55,
	// "1" Bitmap
	0x22, // 0x00, 0x50,	0x00, 0x50,
	0x22, // 0x00, 0x50,	0x00, 0x50,
	0x22, // 0x00, 0x50,	0x00, 0x50,
	0x22, // 0x00, 0x50,	0x00, 0x50,
	// "2" Bitmap
	0xF1, // 0x55, 0x55,	0x00, 0x05,
	0x1F, // 0x00, 0x05,	0x55, 0x55,
	0x88, // 0x50, 0x00,	0x50, 0x00,
	0x8F, // 0x50, 0x00,	0x55, 0x55,
	// "3" Bitmap
	0xF1, // 0x55, 0x55,	0x00, 0x05,
	0x1F, // 0x00, 0x05,	0x55, 0x55,
	0x11, // 0x00, 0x05,	0x00, 0x05,
	0x1F, // 0x00, 0x05,	0x55, 0x55,
	// "4" Bitmap
	0x89, // 0x50, 0x00,	0x50, 0x05,
	0x99, // 0x50, 0x05,	0x50, 0x05,
	0xF1, // 0x55, 0x55,	0x00, 0x05,
	0x11, // 0x00, 0x05,	0x00, 0x05,
	// "5" Bitmap
	0xF8, // 0x55, 0x55,	0x50, 0x00,
	0x8F, // 0x50, 0x00,	0x55, 0x55,
	0x11, // 0x00, 0x05,	0x00, 0x05,
	0x1F, // 0x00, 0x05,	0x55, 0x55,
	// "6" Bitmap
	0x88, // 0x50, 0x00,	0x50, 0x00,
	0x8F, // 0x50, 0x00,	0x55, 0x55,
	0x99, // 0x50, 0x05,	0x50, 0x05,
	0x9F, // 0x50, 0x05,	0x55, 0x55,
	// "7" Bitmap
	0xF3, // 0x55, 0x55,	0x00, 0x55,
	0x11, // 0x00, 0x05,	0x00, 0x05,
	0x11, // 0x00, 0x05,	0x00, 0x05,
	0x11, // 0x00, 0x05,	0x00, 0x05,
	// "8" Bitmap
	0xF9, // 0x55, 0x55,	0x50, 0x05,
	0x9F, // 0x50, 0x05,	0x55, 0x55,
	0x99, // 0x50, 0x05,	0x50, 0x05,
	0x9F, // 0x50, 0x05,	0x55, 0x55,
	// "9" Bitmap
	0xF9, // 0x55, 0x55,	0x50, 0x05,
	0x9F, // 0x50, 0x05,	0x55, 0x55,
	0x11, // 0x00, 0x05,	0x00, 0x05,
	0x11  // 0x00, 0x05,	0x00, 0x05,
	};

//
// Note: The comments are just a guide to help me map
// from an old bitmapping scheme to the new, more densely packed, format
//
// 0 = 00, 1 = 02, 2 = 07, 3 = 03
// 4 = 20, 5 = 22, 6 = 27, 7 = 23
// 8 = 70, 9 = 72, A = 77, B = 73
// C = 30, D = 32, E = 37, F = 33
//
const BYTE Numbers_Normal [80] = {
	// "0" Bitmap
	0x28, // 0x07, 0x70,
	0x82, // 0x70, 0x07,
	0x82, // 0x70, 0x07,
	0x82, // 0x70, 0x07,
	0x82, // 0x70, 0x07,
	0x82, // 0x70, 0x07,
	0x82, // 0x70, 0x07,
	0x28, // 0x07, 0x70,
	// "1" Bitmap
	0x08, // 0x00, 0x70,
	0x28, // 0x07, 0x70,
	0x08, // 0x00, 0x70,
	0x08, // 0x00, 0x70,
	0x08, // 0x00, 0x70,
	0x08, // 0x00, 0x70,
	0x08, // 0x00, 0x70,
	0x08, // 0x00, 0x70,
	// "2" Bitmap
	0xEB, // 0x37, 0x73,
	0x82, // 0x70, 0x07,
	0x82, // 0x70, 0x07,
	0x07, // 0x00, 0x23,
	0x38, // 0x03, 0x70,
	0x2C, // 0x07, 0x30,
	0xD0, // 0x32, 0x00,
	0xAA, // 0x77, 0x77,
	// "3" Bitmap
	0x2B, // 0x07, 0x73,
	0x82, // 0x70, 0x07,
	0x02, // 0x00, 0x07,
	0x39, // 0x03, 0x72,
	0x39, // 0x03, 0x72,
	0x02, // 0x00, 0x07,
	0x82, // 0x70, 0x07,
	0x2B, // 0x07, 0x73,
	// "4" Bitmap
	0x3B, // 0x03, 0x73,
	0x27, // 0x07, 0x23,
	0x77, // 0x23, 0x23,
	0x87, // 0x70, 0x23,
	0x87, // 0x70, 0x23,
	0xA9, // 0x77, 0x72,
	0x0B, // 0x00, 0x73,
	0x0B, // 0x00, 0x73,
	// "5" Bitmap
	0xA9, // 0x77, 0x72,
	0x80, // 0x70, 0x00,
	0x80, // 0x70, 0x00,
	0x97, // 0x72, 0x23,
	0x02, // 0x00, 0x07,
	0x02, // 0x00, 0x07,
	0x02, // 0x00, 0x07,
	0xAB, // 0x77, 0x73,
	// "6" Bitmap
	0xEB, // 0x37, 0x73,
	0x82, // 0x70, 0x07,
	0x80, // 0x70, 0x00,
	0xAB, // 0x77, 0x73,
	0xBE, // 0x73, 0x37,
	0x82, // 0x70, 0x07,
	0x82, // 0x70, 0x07,
	0xEB, // 0x37, 0x73,
	// "7" Bitmap
	0xAA, // 0x77, 0x77,
	0x0E, // 0x00, 0x37,
	0x0D, // 0x00, 0x32,
	0x0B, // 0x00, 0x73,
	0x38, // 0x03, 0x70,
	0x2C, // 0x07, 0x30,
	0xE0, // 0x37, 0x00,
	0xD0, // 0x32, 0x00,
	// "8" Bitmap
	0xEB, // 0x37, 0x73,
	0x82, // 0x70, 0x07,
	0x82, // 0x70, 0x07,
	0xEB, // 0x37, 0x73,
	0xBE, // 0x73, 0x37,
	0x82, // 0x70, 0x07,
	0x82, // 0x70, 0x07,
	0xEB, // 0x37, 0x73,
	// "9" Bitmap
	0x28, // 0x07, 0x70,
	0x82, // 0x70, 0x07,
	0x82, // 0x70, 0x07,
	0x69, // 0x27, 0x72,
	0x2E, // 0x07, 0x37,
	0x02, // 0x00, 0x07,
	0x82, // 0x70, 0x07,
	0xD8  // 0x32, 0x70,
	};

typedef enum {NORMAL, GREEN, WHITE} NumberType;
NumberType Numbers = NORMAL;	// Indicates one or other of the sets of digits

#endif

//
// These values can be overridden on the command line. Format is:
// U=X M=Y T=Z O S D=Y P=PATH
// where "X" is the Interval in seconds and "Y" is MemToFree in MBytes,
// and "Z" is a Minimum Threshold in MBytes at which memory is automatically freed.
// O means run Once - frees memory then exits
// S means run Sound - beeps when freeing memory automatically
// D=Y/N means defragment/don't defragment. Default defragments when mem to free < 1/3 total
// P=PATH specifies the path to RAMpageConfig.exe. MUST be last if specified. (PATH is not quoted.)
//
unsigned int Interval;  // Number of milliseconds between status checks
DWORD MemToFree;        // Number of bytes to attempt to free
DWORD MinMemThreshold;  // Number of bytes below which FreeRAM is triggered automatically
BOOL  bRunOnce = FALSE;	// If TRUE, the program frees memory and exits
BOOL  bBeep = FALSE;	// If TRUE, the program beeps when freeing memory automatically
BOOL  bAlwaysFree = FALSE;	// If TRUE, the the program frees memory even if the last attempt failed
BOOL  bReachTarget = FALSE; // If TRUE, MemToFree represents an absolute target.
unsigned int  GraphType;	// If 1 or 2, the program displays a graph instead of a numeric readout
char  Defrag = 'I';		// 'N' = Never defrag, 'Y' = Always defrag, Any other char = "Intelligent" defrag
LPSTR ConfigPath = NULL;// Directory spec for the configuration screen program.

unsigned int Doubling;		// Number of time-slices until next auto free is allowed, (doubles for each free)
unsigned int MaxDoubling;	// Doubling limit. (Calculated to be 10 minutes based on the selected Interval)
unsigned int Slices;		// Counts the number of time slices since the last auto free.
unsigned int DoublingInitializer;	// Initial value for Doubling - 
									// Higher for low intervals to even out the initial gap between auto frees.

// Note: IconBitmap is used for a number of purposes, (most added after it got its misleading name):
// 1. It stores the icon bitmap (foreground & background) in the non-hidden version.
//    This requires 256 bytes, (2 bitmaps of 128 bytes each.)
// 2. It also stores the Tooltip in the non-hidden version. This is in bytes 128 -> 128+64. 
//    The tooltip is read into the nid buffer before overwriting it with the background bitmap.
// 3. When freeing RAM, (both versions), the bitmap is used to store an array of 16 pointers,
//    (128 bytes), while performing an undefragmented freeing operation.

#ifndef _RAMPAGE_HIDDEN

BYTE IconBitmap[256];
HINSTANCE ghInst;

#else

BYTE IconBitmap[128];

#endif

LRESULT CALLBACK wndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
HWND ghWnd;

#ifndef _RAMPAGE_HIDDEN

//
// Updates the icon in the System Tray
//
void UpdateIcon ()
{
	HDC hdc;
    NOTIFYICONDATA nid;
	HBITMAP XANDhbitmap;
	HBITMAP XORhbitmap;
	ICONINFO iconinfo;
	int firsttime = 0;

	if (IconBitmap[128] == '\0')	// Initializing - Add the sys tray icon
	{
		memcpy(nid.szTip,classname,63); // Copying some extra garbage here - but no harm done
		firsttime = 1;
	}
	else
		memcpy(nid.szTip,IconBitmap+128,63);	// Copying some extra garbage here - but no harm done

#ifdef _RAMPAGE_COLOR_DEBUG
	int i, j;

	for (i=2; i<16; i++)
		for (j=0; j<8; j++)
			IconBitmap[(i*8)+j] = ((j/2)+(4*(i/9))) + (((j/2)+(4*(i/9)))<<4);
#endif

	hdc = GetDC(ghWnd);	// get a handle to the display device context
	//
	// Create a device dependant bitmap (DDB) using our device independant
	// bitmap information, initializing the bitmap with the values in IconBitmap
	//
	memset(IconBitmap+128,BMP_BLACK,128);
	XANDhbitmap = CreateDIBitmap(
		hdc,										// handle to device context 
		&localbmi.bmiHeader,						// pointer to bitmap size and format data 
		CBM_INIT,									// initialization flag 
		IconBitmap+128,								// pointer to initialization data 
		(const struct tagBITMAPINFO *)&localbmi,	// pointer to bitmap color-format data 
		DIB_RGB_COLORS 								// color-data usage 
		);
	XORhbitmap = CreateDIBitmap(
		hdc,										// handle to device context 
		&localbmi.bmiHeader,						// pointer to bitmap size and format data 
		CBM_INIT,									// initialization flag 
		IconBitmap,									// pointer to initialization data 
		(const struct tagBITMAPINFO *)&localbmi,	// pointer to bitmap color-format data 
		DIB_RGB_COLORS 								// color-data usage 
		);

	//
	// Create the icon using the DBB
	//
	if ((XANDhbitmap != NULL) && (XORhbitmap != NULL))
	{
		iconinfo.fIcon = TRUE; 
		iconinfo.xHotspot = 0; 
		iconinfo.yHotspot = 0; 
		iconinfo.hbmMask = XANDhbitmap;
		iconinfo.hbmColor = XORhbitmap; 
		nid.hIcon = CreateIconIndirect(&iconinfo);
	}
	else
		nid.hIcon = NULL;

	if (XANDhbitmap != NULL)
		DeleteObject(XANDhbitmap);
	if (XORhbitmap != NULL)
		DeleteObject(XORhbitmap);

	ReleaseDC(ghWnd, hdc);

	//
	// Load the RAMpage icon if CreateIconIndirect failed
	//
//    if (nid.hIcon == NULL)
//      nid.hIcon = LoadImage(ghInst, MAKEINTRESOURCE(IDI_ICON), IMAGE_ICON,
//        16, 16, 0); // 16x16 icon

	nid.cbSize = sizeof(NOTIFYICONDATA); // size
	nid.hWnd = ghWnd; // window to receive notifications
	nid.uID = 1;      // application-defined ID for icon (can be any UINT value)
	nid.szTip[63] = '\0';  // guarantee that the tooltip is null terminated.

	if (firsttime)	// Initializing - Add the sys tray icon
	{
		memcpy(nid.szTip,classname,63); // Copying some extra garbage here - but no harm done
		nid.uFlags = NIF_MESSAGE |	// nid.uCallbackMessage is valid, use it
					NIF_ICON |		// nid.hIcon is valid, use it
					NIF_TIP;		// nid.szTip is valid, use it
		nid.uCallbackMessage = UWM_SYSTRAY; // message sent to nid.hWnd
		Shell_NotifyIcon(NIM_ADD, &nid);
	}
	else					// Updating - Modify the sys tray icon
    {
		nid.uFlags = NIF_ICON | NIF_TIP; 
		Shell_NotifyIcon(NIM_MODIFY, &nid);
	}

	if (nid.hIcon != NULL)
		DestroyIcon(nid.hIcon);
}

void NybbleShift(const LPBYTE pX, const LPBYTE pN, const unsigned i)
{
	*(pX-(i*8)) = *pN >> 4;
	*(pX-(i*8)+1) = (*pN << 4) | (*(pN+1) >> 4);
	*(pX-(i*8)+2) = (*(pN+1) << 4);
}

LPBYTE GetColorBytes(const LPBYTE ColorBytes, const unsigned digit, const unsigned i)
{

	if (Numbers == NORMAL)
	{
		*ColorBytes = 
			(((BYTE)Numbers_Normal[(digit * 8) + i] & (BYTE)0xC0) >> 2) |
			(((BYTE)Numbers_Normal[(digit * 8) + i] & (BYTE)0x30) >> 4);
		*(ColorBytes+1) = 
			(((BYTE)Numbers_Normal[(digit * 8) + i] & (BYTE)0x0C) << 2) |
			 ((BYTE)Numbers_Normal[(digit * 8) + i] & (BYTE)0x03);
	}
	else
	{
		*ColorBytes = 
			(((BYTE)(Numbers_Luminous[(digit * 4) + (i/2)] << ((i%2)*4)) & (BYTE)0x80) >> 1) |
			(((BYTE)(Numbers_Luminous[(digit * 4) + (i/2)] << ((i%2)*4)) & (BYTE)0x40) >> 4);
		*(ColorBytes+1) = 
			(((BYTE)(Numbers_Luminous[(digit * 4) + (i/2)] << ((i%2)*4)) & (BYTE)0x20) << 1) |
			(((BYTE)(Numbers_Luminous[(digit * 4) + (i/2)] << ((i%2)*4)) & (BYTE)0x10) >> 2);

		if (Numbers == WHITE)
		{
			*ColorBytes >>= 1;
			*(ColorBytes+1) >>= 1;
		}
	}

	return (ColorBytes);
}

#endif

void FreeRAM ();

//
// Updates the current memory status
//
void UpdateRamPageInfo(const BOOL bUpdateIfUnchanged)
{
	static MEMORYSTATUS last_memstat = 
		{(DWORD)-1,(DWORD)-1,(DWORD)-1,(DWORD)-1,(DWORD)-1,(DWORD)-1,(DWORD)-1};
	MEMORYSTATUS memstat;

#ifndef _RAMPAGE_HIDDEN

	unsigned thousands;
	unsigned hundreds;
	unsigned tens;
	unsigned units;
	unsigned i;
	unsigned row;
	unsigned col;
	static unsigned lastpt = 20;
	unsigned thispt;
	BYTE ColorBytes[2];
	BYTE BarColor;

#endif

	//
	// Get Memory Status Information
	//
	memstat.dwLength = sizeof(MEMORYSTATUS);
	GlobalMemoryStatus(&memstat);

#ifdef _RAMPAGE_FREE_DEBUG
	if (memstat.dwAvailPhys <= MinMemThreshold) {
		sprintf(DebugStr,"Free Memory (%d) <= Threshold (%d)!",
			memstat.dwAvailPhys, MinMemThreshold);
		MessageBox(NULL, DebugStr, classname, MB_TASKMODAL);
	}
#endif

	if (MinMemThreshold && !bUpdateIfUnchanged)
	{
		if (Slices == Doubling)
		{
			if ((memstat.dwAvailPhys > MinMemThreshold))
				Doubling = DoublingInitializer;	// reset the doubling factor
			Slices = 0;
		}

		if ((memstat.dwAvailPhys <= MinMemThreshold) && (bAlwaysFree || !Slices))
		{
			//
			// Attempt to free some Physical RAM
			//
			if (bBeep)
				MessageBeep(MB_ICONEXCLAMATION);
			FreeRAM ();
			if (!bAlwaysFree)
			{
				Slices++;
				Doubling *= 2;							// double up
				if (Doubling > MaxDoubling)				// Limit this doubling so that at least 
					Doubling = MaxDoubling;				// 1 attempt is made every 10 minutes
			}
			return;
		}
		else if (Doubling != DoublingInitializer)
			Slices++;
	}

	memstat.dwAvailPhys = (memstat.dwAvailPhys + HALF_MBYTE)  / MBYTE;
	memstat.dwAvailPageFile = (memstat.dwAvailPageFile + HALF_MBYTE) / MBYTE;
    memstat.dwTotalPhys = (memstat.dwTotalPhys + HALF_MBYTE) / MBYTE;

	//
	// Don't do any further processing if the memory status hasn't changed
	// and bUpdateIfUnchanged is FALSE
	//
	if ((last_memstat.dwAvailPhys == memstat.dwAvailPhys) &&
		(last_memstat.dwAvailPageFile == memstat.dwAvailPageFile) &&
		!bUpdateIfUnchanged && !GraphType)
#ifdef _RAMPAGE_DEBUG
	// In debug mode, we update the test patterns every time
	if (!bTesting)
#endif
		return;

	//
	// Remember the memory status for next time
	//
	last_memstat.dwAvailPhys = memstat.dwAvailPhys; 
	last_memstat.dwAvailPageFile = memstat.dwAvailPageFile;

#ifndef _RAMPAGE_HIDDEN

	//
	// Create the ToolTip
	//
#ifdef _RAMPAGE_RESMON
	GetResourceInfo();
	sprintf(IconBitmap+128, TOOLTIP, memstat.dwAvailPhys, memstat.dwAvailPageFile,
		lSysRes,lUsrRes,lGdiRes);
#else
	sprintf(IconBitmap+128, TOOLTIP, memstat.dwAvailPhys, memstat.dwAvailPageFile);
#endif

	//
	// For display purposes, we need to extract each digit from the 
	// total available physical memory.
	//
	thousands = ((memstat.dwAvailPhys % (DWORD)10000) / (DWORD)1000);
	hundreds = ((memstat.dwAvailPhys % (DWORD)1000) / (DWORD)100);
	tens = ((memstat.dwAvailPhys % (DWORD)100) / (DWORD)10);
	units = (memstat.dwAvailPhys % (DWORD)10);
	
	//
	// Figure out what color to make the bar below the numbers.
	//
	// If Total Physical RAM is greater than 999M, and there is less than 1000M
	// available, set the bar color to YELLOW. This helps a little to distinguish 
	// between 1400M and 400M for example, (1400 would show as 400 with a GREEN
	// bar, 400 would show as 400 with a YELLOW bar.
	// 
	if (Numbers == NORMAL)
		BarColor = BMP_DARK_GREEN;
	else
		BarColor = BMP_GREEN;
	if (!GraphType && (memstat.dwTotalPhys % (DWORD)10000) / (DWORD)1000)
	{
		if (!thousands)
			BarColor = BMP_YELLOW;
	}
	else
		//
		// For anything with less than 1000M, the bar goes YELLOW when
		// Free Memory dips below 20% of total.
		//
		if (((memstat.dwAvailPhys*100) / memstat.dwTotalPhys) < 20)
			BarColor = BMP_YELLOW;

	//
	// RED bar appears whenever Free Memory dips below 5% of total
	//
	if (((memstat.dwAvailPhys*100) / memstat.dwTotalPhys) < 5)
		BarColor = BMP_RED;

	if (GraphType)
	{
		thispt = ((memstat.dwAvailPhys*14) + (memstat.dwTotalPhys/28)) / memstat.dwTotalPhys;
		if (!thispt)
			thispt = 1;

		if ((GraphType == 2) || (GraphType == 3))
		{
			if (lastpt == 20)
				lastpt = thispt;

			for (col=0; col<8; col++)
			{
				for (row=8; row<120; row+=8)
				{
					if (col && (IconBitmap[row+col] & 0x0F))
						IconBitmap[row+col] = BarColor << 4;
					else
						IconBitmap[row+col] = 0;
					if (col != 7)
					{
						if (IconBitmap[row+col+1] & 0xF0)
							IconBitmap[row+col] |= BarColor >> 4;
					}
					else
					{
						IconBitmap[row+7] = 0;

						if ( ((GraphType == 2) && (thispt >= (row/8))) || 
							 ((GraphType == 3) && (
								(thispt == (row/8))
								|| ((lastpt < (row/8)) && (thispt > (row/8)))
								|| ((lastpt > (row/8)) && (thispt < (row/8)))
							 )))
							IconBitmap[row+7] |= BarColor << 4;
					}
				}
			}
			lastpt = thispt;
		}
		else
		{
			for (row=8; row<120; row+=8)
				for (col=0; col<8; col++)
					if (thispt >= (row/8))
					{
						IconBitmap[row+col] = BarColor;
						if (!col)
							IconBitmap[row+col] = BarColor >> 4;
						if (col==7)
							IconBitmap[row+col] = BarColor << 4;
					}
					else
						IconBitmap[row+col] = 0;
		}
	}
	else
	{
		//
		// Reset the Icon bitmap.
		//
		memset(IconBitmap,BMP_BLACK,128);

#ifdef _RAMPAGE_DEBUG
		// In debug mode, we don't actually display the free RAM, 
		// we display some test patterns instead
		if (bTesting)
		{
			if (test_num == 0)
				thousands = 1;
			hundreds = test_num;
			tens = test_num;
			units = test_num;
			test_num++;
			if (test_num == 11)
				test_num = 0;
		}

		if (bTesting && (test_num == 0))
			for (i=0; i<16; i++)
				memset(IconBitmap+(8*i),i+(i<<4),8);	// test color card
		else
#endif
		{
			//
			// This section fills in the icon bitmap with the appropriate digits.
			// The bit shifting moves the digits 4 bits, (one pixel), right and left.
			//
			for (i=0; i<8; i++) // digits are 8 pixels high
			{
				if (hundreds || thousands)
				{
					// "hundreds" digit
					NybbleShift(IconBitmap+24+72, GetColorBytes(ColorBytes,hundreds,i), i);

					// "units" digit
					NybbleShift(IconBitmap+29+72, GetColorBytes(ColorBytes,units,i), i);

					// "tens" digit
					memcpy(IconBitmap+27+72-(i*8),GetColorBytes(ColorBytes,tens,i),2);

				}
				else if (tens)
				{
					// "tens" digit
					memcpy(IconBitmap+26+72-(i*8),GetColorBytes(ColorBytes,tens,i),2);

					// "units" digit
					NybbleShift(IconBitmap+28+72, GetColorBytes(ColorBytes,units,i), i);
	
				}
				else
					// "units" digit
					memcpy(IconBitmap+27+72-(i*8),GetColorBytes(ColorBytes,units,i),2);
	
			}

			//
			// Change the color below the digits
			//
			IconBitmap[16] = (BarColor << 4);
			IconBitmap[16] = (IconBitmap[16] >> 4);
			memset(IconBitmap+17,BarColor,6);
			IconBitmap[23] = (BarColor >> 4) << 4;

		}
	}

    UpdateIcon ();

#endif
}

//
// Attempt to free RAM by allocating/freeing memory
//
void FreeRAM ()
{
  void ** pMem;
  MEMORYSTATUS memstat;
  size_t MemSize = 0;
  int i;

  memstat.dwLength = sizeof(MEMORYSTATUS);
  GlobalMemoryStatus(&memstat);
  if (!bReachTarget)
    MemSize = (size_t)(MemToFree + (size_t)(memstat.dwAvailPhys));
  else if ((size_t)(memstat.dwAvailPhys) < MemToFree)
    MemSize = MemToFree;

  if (!MemSize)
	return;

  KillTimer(ghWnd, 0x29a);

#ifndef _RAMPAGE_HIDDEN

  memcpy(IconBitmap+128,ATTEMPTING_TO_FREE_RAM,LEN_ATTEMPTING_TO_FREE_RAM);  // Saves a few bytes of strcpy code!
  if (GraphType)
    memset(IconBitmap,BMP_BLACK,128);
  else
    memset(IconBitmap+32,BMP_BLACK,96);
//  memset(IconBitmap,BMP_BLACK,16);
//  memset(IconBitmap+16,BMP_RED,96);
//  memset(IconBitmap+112,BMP_BLACK,16);
//  for (i=16; i<112; i+=8)
//  {
//	memset(IconBitmap+i,BMP_BLACK,1);
//	memset(IconBitmap+i+7,BMP_BLACK,1);
//  }
  UpdateIcon();

#endif

  //
  // Use IconBitmap as a temporary store for 16 pointers to allocated memory.
  // (If we are not defragmenting, we allocate 16 chunks of memory to try to meet the target.)
  //
  pMem = (void **)IconBitmap;

#ifdef _RAMPAGE_FREE_DEBUG
  sprintf(DebugStr,"Attempting to allocate %d bytes", MemSize);
  MessageBox(NULL, DebugStr, classname, MB_TASKMODAL);
#endif

  if (MemSize > (size_t)(memstat.dwTotalPhys - (8 * MBYTE)))
	  MemSize = (size_t)(memstat.dwTotalPhys - (8 * MBYTE));

  if ((Defrag == 'N') || ((Defrag != 'Y') && (MemSize > (size_t)(memstat.dwTotalPhys / 3))))
  {
	  for (i=0; (i<16) && ((pMem[i] = malloc(MemSize/16)) != NULL ); i++)
		  memset(pMem[i],0,MemSize/16);
	  for (i=0; (i<16) && (pMem[i] != NULL ); i++) 
	  {
	    free(pMem[i]);
		pMem[i] = NULL;
	  }

	  if (Defrag != 'Y')
	  {
		  if (MemSize/16 < MemToFree)	// Always ensure that one chunk is at least as big
			  MemSize = MemToFree;		// as the "MemToFree" parameter specified by the user.
		  else
			  MemSize = 16;				// Allocate a few bytes 
	  }
  }

  //
  // If we were not bothered with defragmenting, MemSize = 16bytes so this next malloc doesn't
  // really do much of anything.
  // If we are not defragmenting because there is a lot of free memory, MemSize is set = 
  // MemToFree to ensure that at least one chunk of memory is the same size as MemToFree.
  // If we are ALWAYS defragmenting, the above block will have been skipped, so MemSize will
  // represent the full amount of memory we are attempting to free up.
  //
  while (((pMem[0] = malloc(MemSize)) == NULL ) && (MemSize > (size_t)20))
  {
    MemSize = MemSize / (size_t)2;
#ifdef _RAMPAGE_FREE_DEBUG
    sprintf(DebugStr,"RETRYING! Attempting to allocate %d bytes", MemSize);
    MessageBox(NULL, DebugStr, classname, MB_TASKMODAL);
#endif

  }

  if (pMem[0] != NULL)
  {
	memset(pMem[0],0,MemSize);
    free(pMem[0]);
#ifdef _RAMPAGE_FREE_DEBUG
    sprintf(DebugStr,"Allocation succeeded for %d bytes", MemSize);
    MessageBox(NULL, DebugStr, classname, MB_TASKMODAL);
#endif
  }

#ifndef _RAMPAGE_HIDDEN
  if (GraphType)
    memset(IconBitmap,BMP_BLACK,128);
#endif

  UpdateRamPageInfo(TRUE);

  SetTimer(ghWnd, 0x29a, Interval, NULL);
}

const char TokenSepChars[] = " =";

typedef WORD (*typFunc) (WORD restype);
// typFunc GetFreeSystemResources;

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR lpszCmdLine, int nCmdShow)
{
  WNDCLASS wc;
  MSG msg;
  LPSTR param;
  LPSTR value;

  //
  // Process the command line
  //
  Interval = 0;	   	   // Number of milliseconds betweem status checks
  MemToFree = 0;       // Number of bytes to attempt to free
  MinMemThreshold = 0; // Number of bytes below which FreeRAM is triggered automatically
  GraphType = 0;
#ifndef _RAMPAGE_HIDDEN

  Numbers = NORMAL;
/* Adds too much overhead
#ifdef _RAMPAGE_GERMAN
  setlocale(LC_ALL,"germany");
#else
#ifdef _RAMPAGE_SPANISH
  setlocale(LC_ALL,"spain");
#else
#ifdef _RAMPAGE_FRENCH
  setlocale(LC_ALL,"france");
#else
#ifdef _RAMPAGE_DUTCH
  setlocale(LC_ALL,"netherlands");
#else
#ifdef _RAMPAGE_TURKISH
  setlocale(LC_ALL,"turkey");
#endif
#endif
#endif
#endif
#endif
*/
#endif

  // MessageBox(NULL, lpszCmdLine, classname, MB_TASKMODAL);

  param = strtok(lpszCmdLine,TokenSepChars);
  while (param != NULL)
  {
		if ((param[0] == 'U') || (param[0] == 'u'))
		{
			value = strtok(NULL,TokenSepChars);
			Interval = atoi(value);
#ifdef _RAMPAGE_FREE_DEBUG
			sprintf(DebugStr,"Update Value: \t%s\nUpdate Interval: \t%d\n",
				value, Interval);
				MessageBox(NULL, DebugStr, classname, MB_TASKMODAL);
#endif

		}
		else if ((param[0] == 'M') || (param[0] == 'm'))
		{
			value = strtok(NULL,TokenSepChars);
			MemToFree = atoi(value);
		}
		else if ((param[0] == 'R') || (param[0] == 'r'))
		{
			value = strtok(NULL,TokenSepChars);
			MemToFree = atoi(value);
			bReachTarget = TRUE;
		}
		else if ((param[0] == 'D') || (param[0] == 'd'))
		{
			value = strtok(NULL,TokenSepChars);
			Defrag = value[0];
			if ((Defrag >= 'a') && (Defrag <= 'z'))
				Defrag += 'A' - 'a';
		}
		else if ((param[0] == 'P') || (param[0] == 'p'))
		{
			ConfigPath = strtok(NULL,"=\"");
		}
		else if ((param[0] == 'T') || (param[0] == 't'))
		{
			value = strtok(NULL,TokenSepChars);
			MinMemThreshold = ((DWORD)atoi(value)) * MBYTE;
		}
		else if ((param[0] == 'O') || (param[0] == 'o'))
		{
			bRunOnce = TRUE;
		}
		else if ((param[0] == 'S') || (param[0] == 's'))
		{
			bBeep = TRUE;
		}
		else if ((param[0] == 'A') || (param[0] == 'a'))
		{
			bAlwaysFree = TRUE;
		}
		else if ((param[0] == 'G') || (param[0] == 'g'))
		{
			value = strtok(NULL,TokenSepChars);
			GraphType = atoi(value);
		}

#ifndef _RAMPAGE_HIDDEN

		else if ((param[0] == 'L') || (param[0] == 'l'))	// "L" - Luminous digits parameter
		{
			if ((param[1] == 'W') || (param[1] == 'w'))		// "LW" - Luminous white digits parameter
				Numbers = WHITE;
			else
				Numbers = GREEN;
		}
		else if ((param[0] == 'B') || (param[0] == 'b'))	// If this parameter is specified the
			bDetailsInBytes = TRUE;							// "Details" box gives mem. values in bytes

#endif

#ifdef _RAMPAGE_DEBUG
		else if ((param[0] == 'X') || (param[0] == 'x'))	// In debug mode this code allows
			bTesting = TRUE;								// a "debug mode" parameter "X"
#endif

		param = strtok(NULL,TokenSepChars);
  }
  if (!Interval)
	  Interval = 5;
  if (!MemToFree)
	  MemToFree = 10;

  if (Interval < 5)
    DoublingInitializer = (5 - Interval) * 4;	// Jump start the "doubling" process if Interval is low
  else
	DoublingInitializer = 4;

  Doubling = DoublingInitializer;
  //
  // MaxDoubling is used to limit the doubling effect so that 1 attempt to free RAM
  // is made every 10 minutes. This is useful for fully automated systems, (servers, etc),
  // where a process may hog memory for a long time, then not release it properly.
  //
  MaxDoubling = 600 / Interval;
  Slices = 0;

#ifdef _RAMPAGE_FREE_DEBUG
  sprintf(DebugStr,"Memory to Free: \t%d\nMemory Threshold: \t%d\nUpdate Interval: \t%d\n",
	  MemToFree, MinMemThreshold/MBYTE, Interval);
  MessageBox(NULL, DebugStr, classname, MB_TASKMODAL);
#endif

  Interval = Interval * 1000;
  MemToFree = MemToFree * MBYTE;

#ifndef _RAMPAGE_HIDDEN

  ghInst = hInstance;
  // Low priority -- don't soak up as many CPU cycles.
  // This won't make a difference unless we are doing background processing
  // in this process.
  // SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS);

  memset(IconBitmap,BMP_BLACK,128);

#endif

  /* Create a window class for the window that receives systray notifications.
     The window will never be displayed.
  */
  memset((void *)&wc,0,sizeof(WNDCLASS));	// Initialize with zeros
  wc.lpfnWndProc = wndProc;
  wc.hInstance = hInstance;
  wc.lpszClassName = classname;
  RegisterClass(&wc);

  // Create window. Note that WS_VISIBLE is not used, and window is never shown
  ghWnd = CreateWindow(classname, classname, WS_POPUP, 0, 0, 0, 0, NULL, NULL, hInstance, NULL);

#ifndef _RAMPAGE_HIDDEN

  IconBitmap[128] = '\0';
  UpdateIcon ();
  UpdateRamPageInfo (TRUE);

#endif

  while (GetMessage(&msg, NULL, 0, 0)) {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }

  return msg.wParam;

}

LRESULT CALLBACK wndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
#ifndef _RAMPAGE_HIDDEN

  POINT pt;
  HMENU hmenu;
  HMENU hpopup;
  unsigned MenuSelection;
  NOTIFYICONDATA nid;
  MEMORYSTATUS memstat;
  DWORD RoundValue;
  DWORD Divisor;
  char MbyteChar;
#ifdef _RAMPAGE_RESMON
  char szMsg[256];
#else
  char szMsg[192];
#endif

#endif

  ghWnd = hwnd;
  switch (message) {
    case WM_CREATE:
      SetTimer(hwnd, 0x29a, Interval, NULL);
	  if (bRunOnce)
	  {
	      FreeRAM ();
		  bRunOnce = PostMessage(hwnd,WM_DESTROY,0,0);
	  }
      return TRUE;

    case WM_TIMER:
      // Update the Icon/Tooltip
      UpdateRamPageInfo(FALSE);
      return TRUE;

    case WM_DESTROY:
#ifndef _RAMPAGE_HIDDEN

	  nid.cbSize = sizeof(NOTIFYICONDATA); // size
	  nid.hWnd = ghWnd; // window to receive notifications
	  nid.uID = 1;      // application-defined ID for icon (can be any UINT value)
	  nid.uFlags = 0;
	  Shell_NotifyIcon(NIM_DELETE, &nid); // This removes the icon

#endif
      KillTimer(hwnd, 0x29a);
      PostQuitMessage(0);
      return TRUE;

#ifndef _RAMPAGE_HIDDEN

    case UWM_SYSTRAY: // We are being notified of mouse activity over the icon
      switch (lParam) {
        case WM_RBUTTONUP: // Let's track a popup menu

		  GetCursorPos(&pt);
		  hmenu = LoadMenu(ghInst, MAKEINTRESOURCE(IDM_CONTEXTMAIN));
		  hpopup = GetSubMenu(hmenu, 0);

          /* SetForegroundWindow and the ensuing null PostMessage is a
             workaround for a Windows 95 bug (see MSKB article Q135788,
             http://www.microsoft.com/kb/articles/q135/7/88.htm, I think).
             In typical Microsoft style this bug is listed as "by design".
             SetForegroundWindow also causes our MessageBox to pop up in front
             of any other application's windows. */
          SetForegroundWindow(hwnd);
          /* We specifiy TPM_RETURNCMD, so TrackPopupMenu returns the menu
             selection instead of returning immediately and our getting a
             WM_COMMAND with the selection. You don't have to do it this way.
          */
          MenuSelection = TrackPopupMenu(hpopup,    // Popup menu to track
                                 TPM_RETURNCMD |    // Return menu code
                                 TPM_RIGHTBUTTON,   // Track right mouse button?
                                 pt.x, pt.y,        // screen coordinates
                                 0,                 // reserved
                                 hwnd,              // owner
                                 NULL);             // LPRECT user can click in
                                                    // without dismissing menu

		  //
		  // Note: In the original version of this code, the calls to DestroyMenu 
		  // and PostMessage were after the switch statement. This caused a slow
		  // memory leak because of a "return" instead of a "break" statement in
		  // the "Change Display" option.
		  // The bug is fixed, but it does demonstrate a point:
		  // Don't hold on to it any longer than you need it!
		  //
          DestroyMenu(hpopup); // Delete loaded menu and reclaim its resources
          DestroyMenu(hmenu); // Delete loaded menu and reclaim its resources

          PostMessage(hwnd, 0, 0, 0); // see above

		  switch (MenuSelection)
          {
            case IDM_EXIT: 
			  DestroyWindow(hwnd); 
			  break;

			//
			// Display the memory details in a message box
			//
            case IDM_DETAILS:
				memstat.dwLength = sizeof(MEMORYSTATUS);
				GlobalMemoryStatus(&memstat);
#ifdef _RAMPAGE_RESMON
				GetResourceInfo();
#endif

				if (bDetailsInBytes)
				{
					RoundValue = 0;
					Divisor = 1;
					MbyteChar = ' ';
				}
				else
				{
					RoundValue = HALF_MBYTE;
					Divisor = MBYTE;
					MbyteChar = 'M';
				}
				    
#ifdef _REVERSE_LANGUAGE
#ifdef _RAMPAGE_RESMON
				sprintf(szMsg, DetailsMessage,
//						MemTxt,
						PhysicalTxt, TotalTxt, (memstat.dwTotalPhys + RoundValue) / Divisor, MbyteChar,
						PhysicalTxt, FreeTxt,  (memstat.dwAvailPhys + RoundValue) / Divisor, MbyteChar,
						PageTxt,     TotalTxt, (memstat.dwTotalPageFile + RoundValue) / Divisor, MbyteChar,
						PageTxt,     FreeTxt,  (memstat.dwAvailPageFile + RoundValue) / Divisor, MbyteChar,
						VirtualTxt,  TotalTxt, (memstat.dwTotalVirtual + RoundValue) / Divisor, MbyteChar,
						VirtualTxt,  FreeTxt,  (memstat.dwAvailVirtual + RoundValue) / Divisor, MbyteChar,
//						ResTxt,
						FreeTxt,  SysTxt,      lSysRes,
						FreeTxt,  UserTxt,     lUsrRes,
						FreeTxt,  GdiTxt,      lGdiRes);
#else
				sprintf(szMsg, DetailsMessage,
						PhysicalTxt, TotalTxt, (memstat.dwTotalPhys + RoundValue) / Divisor, MbyteChar,
						PhysicalTxt, FreeTxt,  (memstat.dwAvailPhys + RoundValue) / Divisor, MbyteChar,
						PageTxt,     TotalTxt, (memstat.dwTotalPageFile + RoundValue) / Divisor, MbyteChar,
						PageTxt,     FreeTxt,  (memstat.dwAvailPageFile + RoundValue) / Divisor, MbyteChar,
						VirtualTxt,  TotalTxt, (memstat.dwTotalVirtual + RoundValue) / Divisor, MbyteChar,
						VirtualTxt,  FreeTxt,  (memstat.dwAvailVirtual + RoundValue) / Divisor, MbyteChar);
#endif
#else
#ifdef _RAMPAGE_RESMON
				sprintf(szMsg, DetailsMessage,
//						MemTxt,
						TotalTxt, PhysicalTxt, (memstat.dwTotalPhys + RoundValue) / Divisor, MbyteChar,
						FreeTxt,  PhysicalTxt, (memstat.dwAvailPhys + RoundValue) / Divisor, MbyteChar,
						TotalTxt, PageTxt,     (memstat.dwTotalPageFile + RoundValue) / Divisor, MbyteChar,
						FreeTxt,  PageTxt,     (memstat.dwAvailPageFile + RoundValue) / Divisor, MbyteChar,
						TotalTxt, VirtualTxt,  (memstat.dwTotalVirtual + RoundValue) / Divisor, MbyteChar,
						FreeTxt,  VirtualTxt,  (memstat.dwAvailVirtual + RoundValue) / Divisor, MbyteChar,
//						ResTxt,
						FreeTxt,  SysTxt,   lSysRes,
						FreeTxt,  UserTxt,  lUsrRes,
						FreeTxt,  GdiTxt,   lGdiRes);
#else
				sprintf(szMsg, DetailsMessage,
						TotalTxt, PhysicalTxt, (memstat.dwTotalPhys + RoundValue) / Divisor, MbyteChar,
						FreeTxt,  PhysicalTxt, (memstat.dwAvailPhys + RoundValue) / Divisor, MbyteChar,
						TotalTxt, PageTxt,     (memstat.dwTotalPageFile + RoundValue) / Divisor, MbyteChar,
						FreeTxt,  PageTxt,     (memstat.dwAvailPageFile + RoundValue) / Divisor, MbyteChar,
						TotalTxt, VirtualTxt,  (memstat.dwTotalVirtual + RoundValue) / Divisor, MbyteChar,
						FreeTxt,  VirtualTxt,  (memstat.dwAvailVirtual + RoundValue) / Divisor, MbyteChar);
#endif
#endif
					// show memory values in bytes in debug mode
//					sprintf(szMsg, DetailsMessage,
//						TotalTxt, PhysicalTxt, memstat.dwTotalPhys, ' ',
//						FreeTxt,  PhysicalTxt, memstat.dwAvailPhys, ' ',
//						TotalTxt, PageTxt,     memstat.dwTotalPageFile, ' ',
//						FreeTxt,  PageTxt,     memstat.dwAvailPageFile, ' ',
//						TotalTxt, VirtualTxt,  memstat.dwTotalVirtual, ' ',
//						FreeTxt,  VirtualTxt,  memstat.dwAvailVirtual, ' ');

//				else
					// show memory values rounded to nearest MByte in normal mode
//					sprintf(szMsg, DetailsMessage,
//						TotalTxt, PhysicalTxt, (memstat.dwTotalPhys + HALF_MBYTE) / MBYTE, 'M',
//						FreeTxt,  PhysicalTxt, (memstat.dwAvailPhys + HALF_MBYTE) / MBYTE, 'M',
//						TotalTxt, PageTxt,     (memstat.dwTotalPageFile + HALF_MBYTE) / MBYTE, 'M',
//						FreeTxt,  PageTxt,     (memstat.dwAvailPageFile + HALF_MBYTE) / MBYTE, 'M',
//						TotalTxt, VirtualTxt,  (memstat.dwTotalVirtual + HALF_MBYTE) / MBYTE, 'M',
//						FreeTxt,  VirtualTxt,  (memstat.dwAvailVirtual + HALF_MBYTE) / MBYTE, 'M');

				MessageBox(hwnd, szMsg, classname, MB_TASKMODAL);
				break;

			//
			// Display version details and some usage notes in a message box
			//
            case IDM_ABOUT:

#ifdef _RAMPAGE_FREE_DEBUG
			  sprintf(DebugStr,"Memory to Free: \t%d\nMemory Threshold: \t%d\nUpdate Interval: \t%d\n",
				MemToFree/MBYTE, MinMemThreshold/MBYTE, Interval/1000);
			  MessageBox(NULL, DebugStr, classname, MB_TASKMODAL);
#else
			  //
			  // If a path to RAMpageConfig.exe was specified on the command line,
			  // run this program. Otherwise, just display a short help/about message.
			  //
			  if (ConfigPath != NULL)
				  ShellExecute(hwnd, NULL, ConfigPath, "", ".", SW_SHOWNORMAL);
			  else
				MessageBox(hwnd, HelpAboutMessage, classname, MB_TASKMODAL);
#endif

              break;

			//
			// Attempt to free some Physical RAM
			//
            case IDM_FREE:
				FreeRAM ();
				break;

			//
		    // Change the display type and update the Icon/Tooltip
			//
// No longer needed with config utility
//		    case IDM_TOGGLE_DISPLAY:
//			  ToggleNumbers();
//		      UpdateRamPageInfo(TRUE);
//			  break;
          }
          break;

        case WM_LBUTTONDBLCLK:

		  //
		  // Attempt to free some Physical RAM
		  //
		  FreeRAM ();

          break;
        // Other mouse messages: WM_MOUSEMOVE, WM_MBUTTONDOWN, etc.
      }
      return TRUE; // I don't think that it matters what you return.

#endif

  }

  return DefWindowProc(hwnd, message, wParam, lParam);
}

